//==================================
// MAKE3D.C - Matt Pietrek 1994
//==================================
#include <windows.h>
#include <toolhelp.h>
#include <string.h>
#include "ctl3d.h"      // It's assumed you have this file already

HTASK GetHTaskFromExeName(LPSTR lpszName);  // Function prototypes
void CallCtl3dFromAnotherTask_ASM(void);
BOOL CALLBACK _export NotifyCallback(WORD wID, DWORD dwData);

DWORD       OriginalTaskCSIP;   // Original return address from Reschedule
HTASK       HTaskToAdd3dTo;
HINSTANCE   HInstToAdd3dTo;
HTASK       HTaskUs;
BOOL        FSuccessfullyInstalledCtl3D = FALSE;

int PASCAL WinMain( HINSTANCE hInstance,
                    HINSTANCE hPrevInstance,
                    LPSTR lpCmdLine,
                    int nCmdShow )
{
    LPSTR lpszProgramName = lpCmdLine;
    FARPROC lpfnMPIThunk, lpfnMPIThunk2;    // MakeProcInstance thunk
    MSG   msg;

    // If the user didn't specify a program name as a command line
    // argument, default to "PROGMAN".
    if ( *lpszProgramName == 0 )
        lpszProgramName = "PROGMAN";

    // Search through the lists of tasks, looking for a task whose
    // module name matches the name passed on the command line.
    HTaskToAdd3dTo = GetHTaskFromExeName(lpszProgramName);

    if ( !HTaskToAdd3dTo )  // If no task found, bail out
    {
        MessageBox(0, "Couldn't find task", 0, MB_OK);
        return 0;
    }

    HTaskUs = GetCurrentTask();

    // Save away the Reschedule return address of the specified task
    OriginalTaskCSIP = TaskGetCSIP( HTaskToAdd3dTo );

    // Use MakeProcInstance thunks, rather than using __loadds, which
    // would limit us to a single instance.
    lpfnMPIThunk = MakeProcInstance( (FARPROC)CallCtl3dFromAnotherTask_ASM,
                                     hInstance );
    lpfnMPIThunk2 = MakeProcInstance( (FARPROC)NotifyCallback, hInstance );

    // Set up the NotifyRegister callback so that we can wait for the
    // other task to exit.
    NotifyRegister(0, (LPFNNOTIFYCALLBACK)lpfnMPIThunk2, NF_NORMAL);

    // In the other task, start the following sequence:
    // Reschedule -> MakeProcInstance thunk -> _CallCtl3dFromAnotherTask_ASM
    // -> CallCtl3dFromAnotherTask()

    TaskSwitch( HTaskToAdd3dTo, (DWORD)lpfnMPIThunk);

    // If we succeeded in installing CTL3D in the other task's context,
    // wait for the other task to finish.  Theoretically this message
    // loop will only get one message, a WM_QUIT posted by the
    // NotifyCallback() routine when it sees the target task exit.
    if ( FSuccessfullyInstalledCtl3D )
    {
        while ( GetMessage(&msg, 0, 0, 0) )
            DispatchMessage(&msg);

        // Decrement the usage count for CTL3D.DLL.  We shouldn't do
        // this in the NotifyRegister callback, because the loader's
        // in a weird state.
        FreeLibrary( GetModuleHandle("CTL3D.DLL") );
    }

    NotifyUnRegister(0);    // Don't need this anymore!

    FreeProcInstance(lpfnMPIThunk);
    FreeProcInstance(lpfnMPIThunk2);

    return 0;
}

void far CallCtl3dFromAnotherTask(void)
{
    // This is the routine called by the assembler stub
    // (_CallCtl3dFromAnotherTask_ASM).  It executes in the context of
    // the task that we want to make use CTL3D.DLL.  We don't have to use
    // __loadds or __export since the assembler stub set up DS beforehand.
    // Also, that that in this routine, DS != SS.

    // Bump up the CTL3D.DLL reference count by 1.  CTL3D.DLL must already
    // be in memory because the main program (MAKE3D.EXE) implicitly
    // links to it.
    if ( LoadLibrary("CTL3D.DLL") < 32 )
        return;

    if ( Ctl3dRegister(HInstToAdd3dTo) )
        FSuccessfullyInstalledCtl3D = Ctl3dAutoSubclass( HInstToAdd3dTo );
}

HTASK GetHTaskFromExeName(LPSTR lpszName)
{
    // Given the base module name of a task (e.g., "CLOCK" or "PROGMAN"),
    // search through the task list looking for a task with a matching
    // module name.  Return the first HTASK of the first task that's found.

    TASKENTRY te;
    BOOL fMoreToGo;

    te.dwSize = sizeof(te);
    fMoreToGo = TaskFirst(&te);

    while ( fMoreToGo )
    {
        if ( lstrcmpi(te.szModule, lpszName) == 0 )
        {
            HInstToAdd3dTo = te.hInst;
            return te.hTask;
        }
        fMoreToGo = TaskNext(&te);
    }
    return 0;
}

BOOL CALLBACK _export NotifyCallback(WORD wID, DWORD dwData)
{
    // TOOLHELP NotifyRegister callback function.  Here, we're looking for
    // the task that we set up CTL3D in to exit.  When we see it exit,
    // unregister the CTL3D callback, then post a WM_QUIT message to the
    // MAKE3D task.  This will let it drop out of its GetMessage loop and
    // terminate.

    if ( (wID == NFY_EXITTASK) && (GetCurrentTask() == HTaskToAdd3dTo) )
    {
        Ctl3dUnregister(HInstToAdd3dTo);
        PostAppMessage(HTaskUs, WM_QUIT, 0, 0);
    }

    return FALSE;
}

#ifdef __BORLANDC__
// Get rid of BC++ exception handling code - besides making the program
// fatter, it adds RC warnings about multi-instance fixup problems.
// See July 1994 Q&A for an explanation of this trick.
void _ExceptInit(void){}
#endif
